from pathlib import Path
import datetime
import socket

from platformdirs import user_data_dir
from OpenSSL import crypto


__docformat__ = "google"


DEFAULT_KEYS_FOLDER = Path(user_data_dir()) / "tlspyo" / "credentials"


def get_default_keys_folder():
    """
    Creates the default credentials directory and returns it.

    Returns:
        pathlib.Path: default credentials directory
    """
    if not DEFAULT_KEYS_FOLDER.exists():
        DEFAULT_KEYS_FOLDER.mkdir(parents=True, exist_ok=True)
    return DEFAULT_KEYS_FOLDER


def generate_tls_credentials(
        folder_path,
        email_address="emailAddress",
        common_name="default",
        subject_alt_name=('DNS:default',),
        country_name="CA",
        locality_name="localityName",
        state_or_province_name="stateOrProvinceName",
        organization_name="organizationName",
        organization_unit_name="organizationUnitName",
        serial_number=0,
        validity_end_in_seconds=10*365*24*60*60):
    """
    Generates a private TLS key and a self-signed TLS certificate in the designed folder.

    Args:
        folder_path (path-like object): path were the files will be created
        email_address (str): your email address
        common_name (str): your hostname
        subject_alt_name (tuple of str): your subject alt name list
        country_name (str): your country code
        locality_name (str): your locality name
        state_or_province_name (str): your state name
        organization_name (str): your organization name
        organization_unit_name (str): your organization unit name
        serial_number (int): the serial number of your certificate
        validity_end_in_seconds (int): seconds until the generated certificate will expire
    """

    key_file = Path(folder_path) / "key.pem"
    cert_file = Path(folder_path) / "certificate.pem"

    k = crypto.PKey()
    k.generate_key(crypto.TYPE_RSA, 4096)
    cert = crypto.X509()

    subject = cert.get_subject()
    subject.commonName = common_name
    subject.emailAddress = email_address
    subject.organizationName = organization_name
    subject.organizationalUnitName = organization_unit_name
    subject.localityName = locality_name
    subject.stateOrProvinceName = state_or_province_name
    subject.countryName = country_name

    cert.set_issuer(subject)
    cert.gmtime_adj_notBefore(0)
    cert.gmtime_adj_notAfter(validity_end_in_seconds)
    cert.set_pubkey(k)
    cert.set_serial_number(serial_number)
    cert.set_version(2)  # for SAN
    cert.add_extensions([
        crypto.X509Extension(b'subjectAltName', False, ','.join(subject_alt_name).encode())
    ])

    cert.sign(k, 'sha512')
    with open(cert_file, "wt") as f:
        f.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert).decode("utf-8"))
    with open(key_file, "wt") as f:
        f.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k).decode("utf-8"))


def credentials_generator_tool(custom=False):
    """
    Helper tool to generate credentials via CLI.

    Args:
        custom (bool): whether to customize the certificate
    """

    folder_path = get_default_keys_folder()
    email_address = "emailAddress"
    common_name = "default"
    subject_alt_name = ["DNS:" + common_name]
    country_name = "CA"
    locality_name = "localityName"
    state_or_province_name = "stateOrProvinceName"
    organization_name = "organizationName"
    organization_unit_name = "organizationUnitName"
    serial_number = 0
    validity_end_in_seconds = 10 * 365 * 24 * 60 * 60

    if custom:

        print(f"=== TLSPYO - TLS credentials generation tool ===")
        print(f"Please fill the following fields (press ENTER fo leave the default as displayed between brackets)")

        print(f"\nCredentials folder [{folder_path}]:")
        inp = input()
        if inp != "":
            folder_path = Path(inp)
        print(folder_path)

        print(f"\nEmail address [{email_address}]:")
        inp = input()
        if inp != "":
            email_address = inp
        print(email_address)

        print(f"\nCommon name (hostname) [{common_name}]:")
        inp = input()
        if inp != "":
            common_name = inp
        print(common_name)

        subject_alt_name = ["DNS:" + common_name]
        print(f"\nSubject alternative name (hostnames, leave empty to stop adding) {subject_alt_name}:")
        inp = input()
        if inp != "":
            subject_alt_name = []
            while inp != "":
                subject_alt_name.append(inp)
                inp = input()
        print(subject_alt_name)

        print(f"\nCountry code [{country_name}]:")
        inp = input()
        if inp != "":
            country_name = inp
        print(country_name)

        print(f"\nLocality name [{locality_name}]:")
        inp = input()
        if inp != "":
            locality_name = inp
        print(locality_name)

        print(f"\nState or province [{state_or_province_name}]:")
        inp = input()
        if inp != "":
            state_or_province_name = inp
        print(state_or_province_name)

        print(f"\nOrganization name [{organization_name}]:")
        inp = input()
        if inp != "":
            organization_name = inp
        print(organization_name)

        print(f"\nOrganization unit [{organization_unit_name}]:")
        inp = input()
        if inp != "":
            organization_unit_name = inp
        print(organization_unit_name)

        print(f"\nCertificate serial number [{serial_number}]:")
        inp = input()
        if inp != "":
            serial_number = inp
        print(serial_number)

        print(f"\nCertificate validity (in seconds) [{validity_end_in_seconds}]:")
        inp = input()
        if inp != "":
            validity_end_in_seconds = int(inp)
        print(datetime.timedelta(seconds=validity_end_in_seconds))

    generate_tls_credentials(folder_path=folder_path,
                             email_address=email_address,
                             common_name=common_name,
                             subject_alt_name=tuple(subject_alt_name),
                             country_name=country_name,
                             locality_name=locality_name,
                             state_or_province_name=state_or_province_name,
                             organization_name=organization_name,
                             organization_unit_name=organization_unit_name,
                             serial_number=serial_number,
                             validity_end_in_seconds=validity_end_in_seconds)

    print(f"Credentials successfully generated in {folder_path}")


def tcp_broadcast_tls_credentials(port, directory=None):
    """
    Starts a server that broadcasts certificate.pem over TCP.

    Args:
        port: port
        directory: directory where to find certificate.pem if not the default credential directory
    """
    folder = get_default_keys_folder() if directory is None else directory
    cert_path = folder / 'certificate.pem'
    if not cert_path.exists():
        print(f"certificate.pem file not found in {folder}, please generate it first.")
        print("You can generate a certificate.pem in the default tlspyo directory with the following command:")
        print("python -m tlspyo --generate")
        return
    else:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as com_srv:
            com_srv.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            com_srv.bind(('', port))
            com_srv.listen()
            print("Starting credentials server, close the terminal to exit.")
            while True:
                try:
                    conn, addr = com_srv.accept()
                    with conn:
                        with open(cert_path, 'rb') as f:
                            conn.sendfile(f)
                        print(f"Sent TLS certificate to {addr} via port {port}.")
                    print(f"Connection closed.")
                except KeyboardInterrupt:
                    break


def tcp_retrieve_tls_credentials(ip, port, directory=None):
    """
    Starts a client that retrieves certificate.pem over TCP.

    The credential-broadcasting server must be launched via tcp_broadcast_tls_credentials on the Relay machine.

    Args:
        ip: ip of the credential-broadcasting server
        port: port of the credential-broadcasting server
        directory: directory where to write certificate.pem if not the default credentials directory
    """
    folder = get_default_keys_folder() if directory is None else directory
    cert_path = folder / 'certificate.pem'
    print(f"Attempting to reach credentials server at address {ip}:{port}")
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.connect((ip, port))
        while True:
            data = s.recv(1024)
            if not data:
                break
    with open(cert_path, 'wb') as f:
        f.write(data)
    print(f"TLS certificate correctly received in {cert_path}")



from argparse import ArgumentParser

from tlspyo.credentials import credentials_generator_tool, tcp_broadcast_tls_credentials, tcp_retrieve_tls_credentials, get_default_keys_folder


if __name__ == "__main__":
    parser = ArgumentParser(description='Generate, broadcast or retrieve TLS credentials for tlspyo.')
    parser.add_argument('--credentials', action='store_true', help='generates and displays default credentials folder')
    parser.add_argument('--generate', action='store_true', help='generates credentials')
    parser.add_argument('--custom', action='store_true', help='custom credentials generator')
    parser.add_argument('--retrieve', action='store_true', help='retrieve TLS certificate from server')
    parser.add_argument('--broadcast', action='store_true', help='start credentials server')
    parser.add_argument('--directory', action='store', type=str, default='', help='custom credentials directory')
    parser.add_argument('--ip', action='store', type=str, default='127.0.0.1', help='IP address of the credentials server')
    parser.add_argument('--port', action='store', type=int, default=7776, help='port of the credentials server')
    arguments = parser.parse_args()
    ip = arguments.ip
    port = arguments.port
    directory = arguments.directory
    if directory == '':
        directory = None

    if arguments.credentials:
        folder = get_default_keys_folder
        print(f"tlspyo credentials default directory: {folder}")

    if arguments.retrieve:
        tcp_retrieve_tls_credentials(ip=ip, port=port, directory=directory)
    elif arguments.generate:
        credentials_generator_tool(custom=arguments.custom)

    if arguments.broadcast:  # possible to generate then broadcast
        tcp_broadcast_tls_credentials(port=port, directory=directory)



import os
import time
from collections import deque

import OpenSSL
from twisted.internet.protocol import Protocol, Factory
from twisted.internet import ssl

from tlspyo.local_protocol_for_server import LocalProtocolForServerFactory
from tlspyo.credentials import get_default_keys_folder
from tlspyo.logs import logger


class ServerProtocol(Protocol):

    def __init__(self, server):
        self._server = server
        self._identifier = None
        self._state = "HANDSHAKE"
        self._buffer = b""
        self._password = self._server.password
        self._len_password = len(bytes(self._password, encoding='utf8'))
        self._header_size = self._server.header_size

    def connectionMade(self):
        assert self._state == "HANDSHAKE", f"Bad state: {self._state}"
        self.send_obj(cmd="HELLO")

    def connectionLost(self, reason):
        logger.info(f"Connection lost: {reason.getErrorMessage()}")
        if self._server.has_client(self._identifier):
            self._server.delete_client(self._identifier)
        assert not self._server.has_client(self._identifier)
        self._identifier = None
        self._state = "DEAD"

    def process_header(self):
        i = self._header_size + self._len_password
        if len(self._buffer) < i:
            return 0, len(self._buffer) + 1, None
        try:
            data_len = int(self._buffer[:self._header_size])
        except ValueError:
            logger.info(f"Invalid request.")
            self._state = "KILLED"
            self.transport.abortConnection()
            return 0, len(self._buffer) + 1, None
        j = i + data_len
        psw = self._buffer[self._header_size:i]
        psw = psw.decode('utf8')
        if psw != self._password:
            logger.info(f"Invalid password: {psw}")
            self._state = "KILLED"
            self.transport.abortConnection()
        return i, j, psw

    def dataReceived(self, data):
        try:
            self._buffer += data
            i = self._header_size + self._len_password
            if len(self._buffer) >= i:
                i, j, psw = self.process_header()
                if self._state == "KILLED":  # invalid password or request
                    return
                while len(self._buffer) >= j:
                    stamp, cmd, dest, obj = self._server.deserializer(self._buffer[i:j])
                    if cmd == 'ACK':
                        try:
                            del self._server.pending_acks[stamp]  # delete pending ACK
                        except KeyError:
                            logger.warning(f"Received ACK for stamp {stamp} not present in pending ACKs.")
                    else:
                        self.send_ack(stamp)  # send ACK
                        if isinstance(dest, str):
                            dest = (dest, )
                        if cmd == "HELLO":
                            groups = obj
                            if isinstance(groups, str):
                                groups = (groups,)
                            if self._server.check_new_client(groups=groups):
                                logger.info(f"New client with groups {groups}.")
                                self._identifier = self._server.add_client(groups=groups, client=self)
                                self._state = "ALIVE"
                                self.retrieve_broadcast()
                            else:
                                self._state = "CLOSED"
                                self.transport.loseConnection()
                        elif self._state == "ALIVE":
                            if cmd == "OBJ":
                                logger.debug(f"Received object from client {self._identifier} for groups {dest}.")
                                self.forward_obj_to_dest(obj=obj, dest=dest)
                            elif cmd == "NTF":
                                logger.debug(f"Received notification from client {self._identifier} for destination {dest}.")
                                self.retrieve_consumables(groups=dest)
                            else:
                                logger.warning(f"Invalid command: {cmd}")
                                self._state = "CLOSED"
                                self.transport.loseConnection()
                    # truncate the processed part of the buffer:
                    self._buffer = self._buffer[j:]
                    i, j, psw = self.process_header()
                    if self._state == "KILLED":
                        return
        except Exception as e:
            logger.warning(f"Killing connection because of unhandled exception: {e}")
            self._state = "KILLED"
            self.transport.abortConnection()
            raise e

    def send_obj(self, cmd='OBJ', obj=None):
        self._server.ack_stamp += 1
        msg = self._server.serializer((self._server.ack_stamp, cmd, obj))
        msg = bytes(f"{len(msg):<{self._header_size}}", 'utf-8') + msg
        self._server.pending_acks[self._server.ack_stamp] = (time.monotonic(), msg)
        self.transport.write(msg)

    def send_ack(self, stamp):
        msg = self._server.serializer((stamp, 'ACK', None))
        msg = bytes(f"{len(msg):<{self._header_size}}", 'utf-8') + msg
        self.transport.write(msg)

    def retrieve_broadcast(self):
        if self._identifier is not None:
            for _, d_group in self._server.group_info.items():
                if self._identifier in d_group['ids']:
                    to_broadcast = d_group['to_broadcast']
                    if to_broadcast is not None:
                        logger.debug("Sending object {to_broadcast} from retrieve broadcast")
                        self.send_obj(cmd='OBJ', obj=to_broadcast)

    def retrieve_consumables(self, groups):
        if self._identifier is not None:
            for origin, n in groups.items():  # for each requested origin and requested number of consumables
                if origin in self._server.group_info.keys():
                    d_group = self._server.group_info[origin]
                    if self._identifier in d_group['ids']:
                        if n > 0:
                            # add pending consumables from this group for this client
                            d_group['pending_consumers'][self._identifier] += n
                            logger.debug(f"Pending consumables for client {self._identifier} from group {origin}: {d_group['pending_consumers'][self._identifier]}")
                            # retrieve available pending consumables
                            self.dispatch_pending_consumables(origin)
                        elif n < 0:
                            # send all available consumables from this group to this client
                            self.send_all_consumables(origin)

    def dispatch_pending_consumables(self, group):
        # retrieve at most n available consumables from this group
        logger.debug(f"Dispatching consumables from group {group}.")
        if group in self._server.group_info.keys():
            d_group = self._server.group_info[group]
            to_consume = d_group['to_consume']
            pending_consumers = d_group['pending_consumers']
            for id in pending_consumers.keys():
                while pending_consumers[id] > 0 and len(to_consume) > 0:
                    pending_consumers[id] -= 1
                    obj = to_consume.popleft()
                    logger.debug(f"Sending a consumable to client {id} from group {group} (remaining: {pending_consumers[id]}).")
                    self._server.to_clients[id].send_obj(cmd='OBJ', obj=obj)
        else:
            logger.warning(f"Group {group} is not registered in the server.")

    def send_all_consumables(self, group):
        # retrieve all available consumables from this group
        if group in self._server.group_info.keys():
            d_group = self._server.group_info[group]
            to_consume = d_group['to_consume']
            pending_consumers = d_group['pending_consumers']
            while len(to_consume) > 0:
                logger.debug(f'Sending a consumable to client {self._identifier} from group {group}.')
                obj = to_consume.popleft()
                self.send_obj(cmd='OBJ', obj=obj)

    def forward_obj_to_dest(self, obj, dest):
        if dest is not None:
            assert isinstance(dest, dict), f"destination is a {type(dest)}; must be a dict."
            for group, value in dest.items():
                if self._server.try_add_group(group):
                    d_g = self._server.group_info[group]
                    if value < 0:
                        # broadcast object to group
                        d_g['to_broadcast'] = obj
                        ids = d_g['ids']
                        for id_cli in ids:
                            logger.debug(f"Sending object from group {group} to identifier {id_cli}.")
                            self._server.to_clients[id_cli].send_obj(cmd='OBJ', obj=obj)
                    elif value > 0:
                        # add object to group's consumables
                        logger.debug(f"Adding {value} copies of the consumable to group {group}.")
                        for _ in range(value):
                            d_g['to_consume'].append(obj)
                        self.dispatch_pending_consumables(group)

    def get_state(self):
        return self._state


class ServerProtocolFactory(Factory):

    protocol = ServerProtocol

    def __init__(self, server):
        self.server = server

    def buildProtocol(self, addr):
        logger.info('Connected.')
        return ServerProtocol(self.server)


class Server:
    def __init__(self,
                 port,
                 password,
                 serializer,
                 deserializer,
                 accepted_groups=None,
                 header_size=10,
                 local_com_port=2097,
                 security="TLS",
                 keys_dir=None):

        self.serializer = serializer
        self.deserializer = deserializer

        self._port = port
        self._local_com_port = local_com_port
        assert self._local_com_port != self._port, f"Internet and local ports are the same ({self._port})."
        self.password = password
        self.header_size = header_size
        self._accepted_groups = accepted_groups
        self.to_clients = {}  # dict of identifiers to protocols toward clients
        self.group_info = {}  # dictionary of group names to dicts of group info
        self._id_cpt = 0
        self.ack_stamp = 0
        self.pending_acks = {}  # this contains copies of sent commands until corresponding ACKs are received
        self._reactor = None
        self._listener = None
        self._security = security
        self._keys_dir = keys_dir

    def run(self):
        """
        Main loop containing the reactor loop.

        This is run in its own process.
        """
        # from twisted.internet.interfaces import IReadDescriptor
        from twisted.internet import reactor

        # Start local communication
        reactor.connectTCP(host='127.0.0.1', port=self._local_com_port, factory=LocalProtocolForServerFactory(self))

        # Start relay server
        factory = ServerProtocolFactory(self)
        if self._security == "TCP":
            logger.info(f"Listening on TCP to port {self._port}")
            reactor.listenTCP(self._port, factory)
        elif self._security == "TLS":
            # Use default keys if none are provided
            private_key = os.path.join(self._keys_dir, 'key.pem') if self._keys_dir is not None else os.path.join(get_default_keys_folder(), 'key.pem')
            self_signed = os.path.join(self._keys_dir, 'certificate.pem') if self._keys_dir is not None else os.path.join(get_default_keys_folder(), 'certificate.pem')
            # Authenticates the server to all potential clients for TLS communication
            try:
                context = ssl.DefaultOpenSSLContextFactory(private_key, self_signed)
            except OpenSSL.SSL.Error:
                raise AttributeError("The provided keys directory could not be found or does not contain the necessary keys. \
                    Make sure that you are providing a correct path, that your private key is named 'private.key' and that your public key is named 'selfsigned.crt'. \
                    You can use the script generate_certificates.py to generate the keys.")
            logger.info(f"Listening on TLS to port {self._port}, with credentials {private_key} and {self_signed}")
            reactor.listenSSL(self._port, factory, context)
        else:
            logger.warning(f"Unsupported connection: {self._security}")
            return

        self._reactor = reactor
        self._reactor.run()  # main Twisted reactor loop
        self._reactor = None  # remove when done

    def add_accepted_group(self, group, max_count=None, max_consumables=None):
        """
        Adds a new group name to accepted group names

        :param group: str: name of the new group
        :param max_count: maximum number of simultaneous clients in this group
        :param max_consumables: max number of consumables for this group
        """
        if self._accepted_groups is None:
            self._accepted_groups = {}
        self._accepted_groups[group] = {'max_count': max_count,
                                        'max_consumables': max_consumables}

    def check_new_client(self, groups):
        """
        Checks whether a client can be added to requested groups.

        :param groups: list of strings: requested groups for this client
        :return authorization: bool: whether the client can be added
        """
        if self._accepted_groups is not None:
            for group in groups:
                if group not in self._accepted_groups.keys():
                    logger.info(f"Invalid group {group}.")
                    return False
                if group in self.group_info.keys():
                    max_count = self._accepted_groups[group]['max_count']
                    if max_count is not None and len(self.group_info[group]['ids']) >= max_count:
                        logger.info(f"Cannot add more clients to group {group}.")
                        return False
        return True

    def add_client(self, groups, client):
        identifier = self._id_cpt
        self._id_cpt += 1
        self.to_clients[identifier] = client
        for group in groups:
            if self.try_add_group(group):
                logger.debug(f"Adding client {identifier} to group {group}.")
                self.group_info[group]['ids'].append(identifier)
                self.group_info[group]['pending_consumers'][identifier] = 0
        return identifier

    def try_add_group(self, group):
        """
        Adds group if valid.

        :param group: string: requested group
        :return success: bool: whether the group is valid
        """
        if self._accepted_groups is not None:
            if group not in self._accepted_groups.keys():
                logger.info(f"Invalid group {group}.")
                return False
        self.add_group(group)
        return True

    def add_group(self, group, max_consumables=None):
        if group not in self.group_info.keys():
            logger.debug(f"Adding group {group} to relay")
            self.group_info[group] = {'ids': [],  # ids of the clients present in this group
                                      'to_broadcast': None,  # object to broadcast
                                      'to_consume': deque(maxlen=max_consumables) if max_consumables is not None else deque(),  # queue of objects to consume
                                      'pending_consumers': {}  # dict mapping client ids to number of remaining consumables to send from this group
                                      }

    def delete_client(self, identifier):
        for group, d_group in self.group_info.items():
            idents = d_group['ids']
            if identifier in idents:
                logger.debug(f"Removing client {identifier} from group {group}.")
                idents.remove(identifier)
                del d_group['pending_consumers'][identifier]
        logger.debug(f"Removing client {identifier} from list of clients.")
        del self.to_clients[identifier]

    def has_client(self, identifier):
        return identifier in self.to_clients

    def check_acks(self):
        """Returns true if we are not waiting for acknowledgements.

        Returns:
            bool: Whether the dictionary of pending acknowledgements is empty.
        """
        res = len(self.pending_acks.keys()) == 0 
        return res

    def close(self, counter):

        # Check if we are allowed to leave by looking at acknowledgements
        logger.debug(f"Attempting to terminate Relay for {counter}th time")

        if self.check_acks() or counter > 10:
            if self._reactor is not None:
                identifiers = list(self.to_clients.keys())
                for identifier in identifiers:
                    self.to_clients[identifier].transport.loseConnection()
                    self.delete_client(identifier)
                logger.info(f"Succesfully terminated relay connections")
                self._reactor.stop()
        else:
            from twisted.internet import reactor
            reactor.callLater(1, self.close, counter+1)
        

if __name__ == "__main__":
    pass



import queue
from socket import AF_INET, SOCK_STREAM, SOL_SOCKET, SO_REUSEADDR, socket
import pickle as pkl
from threading import Thread, Lock
from multiprocessing import Process
import os

from tlspyo.server import Server
from tlspyo.client import Client

from tlspyo.utils import get_from_queue


__docformat__ = "google"


DEFAULT_SECURITY = "TLS"
DEFAULT_SERIALIZER = pkl.dumps
DEFAULT_DESERIALIZER = pkl.loads


class Relay:
    def __init__(self,
                 port: int,
                 password: str,
                 accepted_groups=None,
                 local_com_port: int = 2096,
                 header_size: int = 10,
                 security: str = DEFAULT_SECURITY,
                 keys_dir: str = None,
                 serializer=None,
                 deserializer=None):
        """
        ``tlspyo`` Relay.

        Endpoints connect to a central Relay, which allows them to communicate in `tlspyo`.
        When used on the Internet, the machine running the Relay must be directly visible from the Internet.
        (Usually, if the machine is behind an Internet box/router, this involves port forwarding.)
        WHEN USING `tlspyo` ON THE INTERNET, IT IS IMPORTANT TO USE TLS SECURITY (default).
        In particular, you will want to choose a strong password, and use your own TLS credentials.
        See the Command Line Interface section of the documentation to generate your TLS credentials.

        Args:
            port (int): port of the Relay
            password (str): password of the Relay
            accepted_groups (object): groups accepted by the Relay.
                If None, the Relay accepts any group;
                Else, must be a dictionary where keys are groups and values are dictionaries with the following entries:

                    - 'max_count': max number of connected clients in the group (None for unlimited)
                    - 'max_consumables': max number of pending consumables in the group (None for unlimited)

            local_com_port (int): local port used for internal communication with Twisted.
            header_size (int): bytes to read at once from socket buffers (the default should work for most cases)
            security (str): one of (None, "TLS");
                None disables TLS, do not use None on a public network unless you know what you are doing!
            serializer (callable): custom serializer that outputs a bytestring from a python object
            deserializer (callable): custom deserializer that outputs a python object from a bytestring
        """

        assert security in (None, "TLS"), f"Unsupported security: {security}"

        if security is None:
            security = "TCP"
        elif security == "SSL":
            security = "TLS"

        assert accepted_groups is None or isinstance(accepted_groups, dict), "Invalid format for accepted_groups."

        self._header_size = header_size
        self._local_com_port = local_com_port
        self._local_com_srv = socket(AF_INET, SOCK_STREAM)
        self._local_com_srv.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)
        self._local_com_srv.bind(('127.0.0.1', self._local_com_port))
        self._local_com_srv.listen()

        keys_dir = os.path.abspath(keys_dir) if keys_dir is not None else keys_dir
        serializer = serializer if serializer is not None else DEFAULT_SERIALIZER
        deserializer = deserializer if deserializer is not None else DEFAULT_DESERIALIZER

        self._server = Server(port=port,
                              password=password,
                              serializer=serializer,
                              deserializer=deserializer,
                              accepted_groups=accepted_groups,
                              local_com_port=local_com_port,
                              header_size=header_size,
                              security=security,
                              keys_dir=keys_dir)
        self._p = Process(target=self._server.run, args=())
        self._p.start()
        self._local_com_conn, self._local_com_addr = self._local_com_srv.accept()
        self._send_local('TEST')

        self._stop_lock = Lock()
        self._stopped = False

    def __del__(self):
        self.stop()

    def _send_local(self, cmd):
        msg = self._server.serializer((cmd, None))
        msg = bytes(f"{len(msg):<{self._header_size}}", 'utf-8') + msg
        self._local_com_conn.sendall(msg)

    def stop(self):
        """
        Stop the Relay.
        """
        try:
            with self._stop_lock:
                if not self._stopped:
                    self._send_local('STOP')

                    self._p.join()
                    self._local_com_conn.close()
                    self._local_com_srv.close()
                    self._local_com_addr = None
                    self._stopped = True
        except KeyboardInterrupt as e:
            self.stop()
            raise e


class Endpoint:
    def __init__(self,
                 ip_server: str,
                 port: int,
                 password: str,
                 groups=None,
                 local_com_port: int = 2097,
                 header_size: int = 10,
                 max_buf_len: int = 4096,
                 security: str = DEFAULT_SECURITY,
                 keys_dir: str = None,
                 hostname: str = "default",
                 serializer=None,
                 deserializer=None,
                 recon_max_delay=60.0,
                 recon_initial_delay=10.0,
                 recon_factor=1.5,
                 recon_jitter=0.1,
                 deserializer_mode="asynchronous"):
        """
        ``tlspyo`` Endpoint.

        Endpoints in tlspyo are python objects that can securely send and receive Python object over the network.
        They communicate via the Relay, and should use TLS security (default) on public networks.
        See the Command Line Interface section of the documentation to generate your TLS credentials.

        Args:
            ip_server (str): the IP address of the Relay (set to '127.0.0.1' for local testing)
            port (int): the port of the Relay (use the same number for the Relay, it must be > 1024)
            password (str): password of the Relay (use the same for the Relay, the stronger, the better)
            groups (tuple of str, or str): groups in which this Endpoint is
            local_com_port (int): local port used for internal communication with Twisted
            header_size (int): number of bytes used for the header (the default should be OK for most cases)
            max_buf_len (int): max bytes to read at once from socket buffers (the default should be OK for most cases)
            security (str): one of (None, "TLS");
                None disables TLS, do not use None on a public network unless you know what you are doing!
            serializer (callable): custom serializer that outputs a bytestring from a python object
            deserializer (callable): custom deserializer that outputs a python object from a bytestring
            recon_max_delay (float): in case of network failure, maximum delay between reconnection attempts
            recon_initial_delay (float): in case of network failure, initial delay between reconnection attempts
            recon_factor (float): in case of network failure, delay will increase by this factor between attempts
            recon_jitter (float): in case of network failure, jitter factor of the delay between attempts
            deserializer_mode (str): one of ("synchronous", "asynchronous"); ("sync", "async") are also accepted;
                in asynchronous mode, objects are deserialized by the receiver thread as soon as they arrive, such that
                they become available to the calling thread as soon as it needs to retrieve them;
                in synchronous mode, objects are deserialized by the calling thread upon object retrieval;
                synchronous mode removes the need for potentially useless, randomly timed deserialization in the
                background, at the cost of performing deserialization upon object retrieval instead
        """

        assert security in (None, "TLS"), f"Unsupported security: {security}"

        if security is None:
            security = "TCP"
        elif security == "SSL":
            security = "TLS"

        # threading for local object receiving
        self.__obj_buffer = queue.Queue()
        self.__socket_closed_lock = Lock() 
        self.__socket_closed_flag = False
    
        # networking (local and internet)
        if isinstance(groups, str):
            groups = (groups, )
        self._header_size = header_size
        self._max_buf_len = max_buf_len
        self._local_com_port = local_com_port
        self._local_com_srv = socket(AF_INET, SOCK_STREAM)
        self._local_com_srv.setsockopt(SOL_SOCKET, SO_REUSEADDR, 1)

        self._deserialize_locally = deserializer_mode in ("synchronous", "sync")

        keys_dir = os.path.abspath(keys_dir) if keys_dir is not None else keys_dir
        serializer = serializer if serializer is not None else DEFAULT_SERIALIZER
        deserializer = deserializer if deserializer is not None else DEFAULT_DESERIALIZER

        self._client = Client(ip_server=ip_server,
                              port_server=port,
                              password=password,
                              serializer=serializer,
                              deserializer=deserializer,
                              groups=groups,
                              local_com_port=local_com_port,
                              header_size=header_size,
                              security=security,
                              keys_dir=keys_dir,
                              hostname=hostname,
                              recon_max_delay=recon_max_delay,
                              recon_initial_delay=recon_initial_delay,
                              recon_factor=recon_factor,
                              recon_jitter=recon_jitter)

        # start local server and Twisted process
        self._local_com_srv.bind(('127.0.0.1', self._local_com_port))
        self._local_com_srv.listen()
        self._p = Process(target=self._client.run, args=())
        self._p.start()
        self._local_com_conn, self._local_com_addr = self._local_com_srv.accept()
        self._send_local(cmd='TEST')

        self._t_manage_received_objects = Thread(target=self._manage_received_objects, daemon=True)
        self._t_manage_received_objects.start()

        self._stop_lock = Lock()
        self._stopped = False

    def __del__(self):
        self.stop()

    def _deserialize(self, obj):
        return self._client.deserializer(obj)

    def _manage_received_objects(self):
        """
        Called in its own thread.
        """
        buf = b""
        while True:
            # Check if socket is still open
            with self.__socket_closed_lock:
                if self.__socket_closed_flag:
                    return

            buf += self._local_com_conn.recv(self._max_buf_len)
            i, j = self._process_header(buf)
            while j <= len(buf):
                stamp, cmd, obj = self._deserialize(buf[i:j])
                if cmd == "OBJ":
                    to_put = obj if self._deserialize_locally else self._deserialize(obj)
                    self.__obj_buffer.put(to_put)  # TODO: maxlen
                buf = buf[j:]
                i, j = self._process_header(buf)

    def _process_header(self, buf):
        i = self._header_size
        if len(buf) < i:
            return 0, len(buf) + 1
        data_len = int(buf[:self._header_size])
        j = i + data_len
        return i, j

    def _send_local(self, cmd, dest=None, obj=None):
        msg = self._client.serializer((cmd, dest, self._client.serializer(obj)))
        msg = bytes(f"{len(msg):<{self._header_size}}", 'utf-8') + msg
        self._local_com_conn.sendall(msg)

    def send_object(self, obj, destination):
        """
        Either broadcast object to destination group(s) or send it as a consumable.

        obj can be any picklable python object.

        destination can either be:
            - a string (single group)
            - a tuple of strings (set of groups)
            - a dictionary where keys are strings (group) and values are integers (number of copies for the group)

        When destination is a string or a tuple of strings, the object will be broadcast to corresponding group(s).

        When destination is a dictionary, each key is a destination group.

        For each corresponding value:
            - if the value is N < 0, the object is broadcast to the group
            - if the value is N > 0, N objects are sent to the group to be consumed. To consume an object, the Endpoint first needs to signal itself as idle for the corresponding group with Endpoint.notify().

        Args:
            obj (object): object to broadcast to destination
            destination (object): destination group(s)
        """
        if isinstance(destination, str):
            destination = {destination: -1}
        elif isinstance(destination, tuple) or isinstance(destination, list):
            destination = {dest: -1 for dest in destination}
        else:
            assert isinstance(destination, dict), f"destination must be either of: str, (str), dict."
            for k, v in destination.items():
                assert isinstance(k, str), f"destination keys must be strings."
                assert isinstance(v, int), f"destination values must be integers."
        assert len(destination.keys()) > 0, f"Please specify at least one group to be notified"
        self._send_local(cmd='OBJ', dest=destination, obj=obj)

    def produce(self, obj, group):
        """
        Alias for send_object(obj=obj, destination={group: 1}).

        Args:
            obj (object): object to send as consumable
            group (str): target group
        """
        assert isinstance(group, str), f"group must be a string, not {type(group)}"
        self.send_object(obj=obj, destination={group: 1})

    def broadcast(self, obj, group):
        """Alias for send_object(obj=obj, destination={group: -1})

        Note that broadcasting an object overrides the previous brodcast object

        Args:
            obj (object): object to send to be broadcast to entire group
            group (str): destination group to which the object should be broadcast.
        """
        assert isinstance(group, str), f"group must be a string, not {type(group)}"
        self.send_object(obj=obj, destination={group: -1})

    def notify(self, groups):
        """
        Notifies the Relay that the Endpoint is ready to retrieve consumables from destination groups.

        groups can either be:
            - a string (single group)
            - a tuple of strings (set of groups)
            - a dictionary where keys are strings (group) and values are integers (number of consumables from the group)

        When groups is a string or a tuple of strings, 1 consumable will be retrieved per corresponding group(s).
        When groups is a dictionary, each key is a destination group.

        For each corresponding value:
            - if the value is N < 0, all available consumables are retrieved from the group
            - if the value is N > 0, at most N available consumables are retrieved from the group

        In any case, the group strings must be a subset of the Endpoint's groups.

        Args:
            groups (object): destination groups of the consumables.
        """
        if isinstance(groups, str):
            groups = {groups: 1}
        elif isinstance(groups, tuple) or isinstance(groups, list):
            groups = {dest: 1 for dest in groups}
        else:
            assert isinstance(groups, dict), f"groups must be either of: str, (str), dict."
            for k, v in groups.items():
                assert isinstance(k, str), f"groups keys must be strings."
                assert isinstance(v, int), f"groups values must be integers."
        assert len(groups.keys()) > 0, f"Please specify at least one group to be notified"
        self._send_local(cmd='NTF', dest=groups, obj=None)

    def stop(self):
        """
        Stop the Endpoint.
        """
        try:
            with self._stop_lock:
                if not self._stopped:
                    # send STOP to the local server
                    self._send_local(cmd='STOP', dest=None, obj=None)

                    # Join the message reading thread
                    with self.__socket_closed_lock:
                        self.__socket_closed_flag = True
                    self._t_manage_received_objects.join()

                    # join Twisted process and stop local server
                    self._p.join()

                    self._local_com_conn.close()
                    self._local_com_srv.close()
                    self._local_com_addr = None
                    self._stopped = True
        except KeyboardInterrupt as e:
            self.stop()
            raise e

    def _process_received_list(self, received_list):
        if self._deserialize_locally:
            for i, obj in enumerate(received_list):
                received_list[i] = self._deserialize(obj)
        return received_list

    def receive_all(self, blocking=False):
        """
        Returns all received objects in a list, from oldest to newest.

        Args:
            blocking (bool): If True, the call blocks until objects are available. Otherwise, the list may be empty.

        Returns:
            list: received objects
        """
        cpy = []
        elem = get_from_queue(self.__obj_buffer, blocking)
        while len(elem) > 0:
            cpy += elem
            elem = get_from_queue(self.__obj_buffer, blocking=False)
        cpy = self._process_received_list(cpy)
        return cpy

    def pop(self, max_items=1, blocking=False):
        """
        Returns at most max_items oldest received objects (FIFO).

        Items are returned from older to more recent.

        Args:
            max_items (int): max number of retrieved items.
            blocking (bool): If True, the call blocks until at least 1 item is retrieved.
                Otherwise, the returned list may be empty.

        Returns:
            list: returned items.
        """
        assert max_items > 0, "Value of max_items must be > 0"

        cpy = []
        elem = get_from_queue(self.__obj_buffer, blocking)
        assert len(elem) == 0 or blocking, 'Issue in pop'
        while len(elem) > 0:
            cpy += elem
            if len(cpy) >= max_items:
                break
            elem = get_from_queue(self.__obj_buffer, blocking=False)
        cpy = self._process_received_list(cpy)
        return cpy

    def get_last(self, max_items=1, blocking=False):
        """
        Returns at most the max_items most recently received objects, and clears receiving buffer.
       
        Items are returned from older to more recent.
        Calling this method clears the receiving buffer.
        In other words, only the most recent objects are retrieved, older objects in the buffer are deleted.
        In case this behavior is not desirable in your application, use `receive_all` or `pop` instead.

        Args:
            max_items (int): maximum number of items to return
            blocking (bool): If True, the call blocks until at least one item is retrieved.
                Otherwise, the returned list may be empty.

        Returns:
            list: The returned items.
        """
        cpy = []
        elem = get_from_queue(self.__obj_buffer, blocking)
        assert len(elem) != 0 or not blocking, 'Issue in get_last'
        while len(elem) > 0:
            cpy += elem
            elem = get_from_queue(self.__obj_buffer, blocking=False)
        cpy = self._process_received_list(cpy[-max_items:])
        return cpy



import unittest
import time
import pickle as pkl

from utils import HelperTester


def same_lists_no_order(l1, l2):
    for elt in l1:
        if elt in l2:
            l2.remove(elt)
        else:
            return False
    return len(l2) == 0


def custom_serializer(obj):
    return b"header" + pkl.dumps(["TEST", pkl.dumps(obj)])


def custom_deserializer(bytestring):
    assert len(bytestring) > len(b"header")
    assert bytestring[:len(b"header")] == b"header"
    bytestring = bytestring[len(b"header"):]
    tmp = pkl.loads(bytestring)
    assert isinstance(tmp, list)
    assert len(tmp) == 2
    assert tmp[0] == "TEST"
    obj = pkl.loads(tmp[1])
    return obj


class TestSerialization(unittest.TestCase):

    # Set up the server and all endpoints for all tests
    def setUp(self):
        self.ht = HelperTester(serializer=custom_serializer, deserializer=custom_deserializer)

    def test_groups_accept_all(self):
        sr = self.ht.spawn_relay
        se = self.ht.spawn_endpoint
        relay = sr(accepted_groups=None)
        ep1 = se(groups='group1')
        ep2 = se(groups='group2')
        time.sleep(1.0)  # let everyone handshake the relay so that broadcasts don't get overwritten before that

        # broadcasting

        ep1.send_object(obj='test1', destination='group2')
        r = ep2.pop(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test1', f"r:{r}")

    def tearDown(self):
        self.ht.clear()


if __name__ == '__main__':
    unittest.main()



import time
import unittest
from utils import HelperTester

NUM_OBJECTS = 10


class TestAPI(unittest.TestCase):

    # Set up the server and all endpoints for all tests
    def setUp(self):
        self.ht = HelperTester()

    def test_read_with_pop(self):
        '''
        Checks pop functionality with broadcast
        '''
        sr = self.ht.spawn_relay
        se = self.ht.spawn_endpoint
        _ = sr(accepted_groups=None) # Starts a relay that accepts all groups
        cons = se(groups='group2')
        prod = se(groups='group1')

        # Checks for weird asynchronous behaviour when nothing has been sent yet
        res = cons.pop(blocking=False)
        self.assertEqual(len(res), 0)        

        # Sends objects to a consumer 
        for i in range(NUM_OBJECTS):
            prod.produce(f"object {i}", group='group2')

        # Check that objects are received and popped in the right order
        cons.notify(groups={'group2': 10})

        for i in range(NUM_OBJECTS):
            res = cons.pop(max_items=1, blocking=True)
            self.assertEqual(res[0], f"object {i}")

        # Checks for weird values for max_items
        self.assertRaises(AssertionError, lambda : cons.pop(max_items=-1, blocking=False))
        self.assertRaises(AssertionError, lambda : cons.pop(max_items=0, blocking=False))

    def test_produce_notify_broadcast(self):
        sr = self.ht.spawn_relay
        se = self.ht.spawn_endpoint
        _ = sr(accepted_groups=None) # Starts a relay that accepts all groups
        prod1 = se(groups='prod1')
        cons1 = se(groups='cons1')
        prod2 = se(groups='prod2')
        cons2 = se(groups='cons2')

        # Check that productions are received in right order by one consumer
        cons1.notify(groups={'cons1': 2})        
        prod1.produce("I'M FIRST", group='cons1')
        prod1.produce("I'M SECOND", group='cons1')
        res = [cons1.pop(blocking=True)[0] for _ in range(2)]
        self.assertEqual(["I'M FIRST", "I'M SECOND"], res)

        # Check that there is no crossover and that elements are sent to right groups
        cons1.notify("cons1")
        cons2.notify("cons2")
        prod1.produce("I'M FOR CONS1", group='cons1')
        prod2.produce("I'M FOR CONS2", group='cons2')
        res1 = cons1.pop(blocking=True)
        res2 = cons2.pop(blocking=True)
        self.assertEqual(res1[0], "I'M FOR CONS1")
        self.assertEqual(res2[0], "I'M FOR CONS2")

        # Check that a broadcast object is received on the other end
        prod1.broadcast("TEST", "cons1")
        res = cons1.pop(blocking=True)
        self.assertEqual(res[0], "TEST")

        # Check for inputs of notify
        self.assertRaises(AssertionError, lambda: cons1.notify(()))
        self.assertRaises(AssertionError, lambda: cons1.notify([]))
        self.assertRaises(AssertionError, lambda: cons1.notify({"": 0.5}))
        self.assertRaises(AssertionError, lambda: cons1.notify({3: "group3"}))
        self.assertRaises(AssertionError, lambda: cons1.notify(42))
        self.assertRaises(AssertionError, lambda: cons1.notify({}))

        # Check for inputs of broadcast
        self.assertRaises(AssertionError, lambda: prod1.broadcast("TEST", 0))
        self.assertRaises(AssertionError, lambda: prod1.broadcast("TEST", 0.5))
        self.assertRaises(AssertionError, lambda: prod1.broadcast("TEST", {}))
        self.assertRaises(AssertionError, lambda: prod1.broadcast("TEST", ()))

        # Check for inputs of produce
        self.assertRaises(AssertionError, lambda: prod1.produce("TEST", 0))
        self.assertRaises(AssertionError, lambda: prod1.produce("TEST", 0.5))
        self.assertRaises(AssertionError, lambda: prod1.produce("TEST", {}))
        self.assertRaises(AssertionError, lambda: prod1.produce("TEST", ()))

        # Check for inputs of send_object
        self.assertRaises(AssertionError, lambda: prod1.send_object("TEST", ()))
        self.assertRaises(AssertionError, lambda: prod1.send_object("TEST", []))
        self.assertRaises(AssertionError, lambda: prod1.send_object("TEST", {"": 0.5}))
        self.assertRaises(AssertionError, lambda: prod1.send_object("TEST", {3: "group3"}))
        self.assertRaises(AssertionError, lambda: prod1.send_object("TEST", 42))
        self.assertRaises(AssertionError, lambda: prod1.send_object("TEST", {}))

    def test_read_with_receive(self):
        '''
        Checks receive all functionality with broadcast.
        '''
        sr = self.ht.spawn_relay
        se = self.ht.spawn_endpoint
        _ = sr(accepted_groups=None) # Starts a relay that accepts all groups
        prod = se(groups='group1')
        cons = se(groups='group2')

        # Tests that no object are received before they are sent (Weird asynchronous behaviour)
        res = cons.receive_all(blocking=False)
        self.assertEqual(len(res), 0)

        cons.notify(groups={"group2": 10})
        # Producer sends objects
        for i in range(NUM_OBJECTS):
            prod.produce(f"object {i}", 'group2')

        # Tests that we have received all objects after a reasonable delay
        res = cons.receive_all(blocking=True)
        # Checks that we have received all objects 
        self.assertEqual(len(res), NUM_OBJECTS)

        # Checks that we have received all objects in the right order
        for i in range(NUM_OBJECTS):
            self.assertEqual(res[i], f"object {i}")

    def test_read_with_get_last(self):
        '''
        Checks get_last functionality with broadcast.
        '''
        sr = self.ht.spawn_relay
        se = self.ht.spawn_endpoint
        _ = sr(accepted_groups=None) # Starts a relay that accepts all groups
        prod = se(groups='group1')
        cons = se(groups='group2')

        # Producer sends objects
        for i in range(NUM_OBJECTS):
            prod.broadcast(f"object {i}", 'group2')

        # Tests that we have received all objects after a reasonable delay
        res = cons.get_last(max_items=5, blocking=True)

        # Checks that we have received all objects in the right order
        self.assertEqual(res[-1], "object 9")

    def tearDown(self):
        self.ht.clear()


if __name__ == '__main__':
    unittest.main(verbosity=2)



import logging

DEFAULT_FORMAT_STR = "%(levelname)s:%(asctime)s:%(message)s"
DEFAULT_LEVEL = logging.getLogger().getEffectiveLevel()

logger = logging.getLogger('tlspyo')
if not logger.handlers:
    console_handler = logging.StreamHandler()
    default_formatter = logging.Formatter(DEFAULT_FORMAT_STR)
    console_handler.setFormatter(default_formatter)
    logger.addHandler(console_handler)
if logger.level == logging.NOTSET:
    logger.setLevel(DEFAULT_LEVEL)
logger.propagate = False



import time
import os

import OpenSSL
from twisted.python.filepath import FilePath
from twisted.internet import ssl
from twisted.internet.protocol import Protocol, ReconnectingClientFactory

from tlspyo.local_protocol_for_client import LocalProtocolForClientFactory
from tlspyo.credentials import get_default_keys_folder
from tlspyo.logs import logger


class ClientProtocol(Protocol):
    def __init__(self,
                 client,
                 password,
                 header_size=10,
                 groups=("default", )):

        self._password = password
        self._header_size = header_size
        self._buffer = b""
        self._groups = groups
        self._client = client
        self._state = "HANDSHAKE"

    def connectionMade(self):
        assert self._state == "HANDSHAKE", f"Bad state: {self._state}"
        self._client.to_server = self

    def connectionLost(self, reason):
        self._state = "DEAD"
        self._client.to_server = None

    def process_header(self):
        i = self._header_size
        if len(self._buffer) < i:
            return 0, len(self._buffer) + 1
        data_len = int(self._buffer[:self._header_size])
        j = i + data_len
        return i, j

    def dataReceived(self, data):
        self._buffer += data
        if len(self._buffer) >= self._header_size:
            i, j = self.process_header()
            while len(self._buffer) >= j:
                stamp, cmd, obj = self._client.deserializer(self._buffer[i:j])
                if cmd == 'ACK':
                    try:
                        logger.debug(f"ACK received after {time.monotonic() - self._client.pending_acks[stamp][0]}s.")
                        del self._client.pending_acks[stamp]  # delete pending ACK
                    except KeyError:
                        logger.warning(f"Received ACK for stamp {stamp} not present in pending ACKs.")
                else:
                    self.send_ack(stamp)  # send ACK
                    if cmd == "HELLO":
                        self.send_obj(cmd='HELLO', obj=self._groups)
                        self._state = "ALIVE"
                        while len(self._client.store) > 0:
                            # send buffered commands to the server
                            cmd, dest, obj = self._client.store[0]
                            self.send_obj(cmd=cmd, dest=dest, obj=obj)  # send buffered command
                            self._client.store = self._client.store[1:]  # remove buffered command from store
                    else:
                        if self._state != "ALIVE":
                            logger.warning(f"Received a command in a bad state: {self._state}.")
                        if cmd == "OBJ":
                            logger.debug(f"Received object, transferring to local EndPoint.")
                            # transfer the object to the EndPoint server
                            if self._client.endpoint is not None:
                                self._client.endpoint.transport.write(self._buffer[:j])
                            else:
                                logger.warning(f"Local EndPoint is not connected, discarding object.")
                # truncate the processed part of the buffer:
                self._buffer = self._buffer[j:]
                i, j = self.process_header()

    def send_obj(self, cmd='OBJ', dest=None, obj=None):
        self._client.ack_stamp += 1
        msg = self._client.serializer((self._client.ack_stamp, cmd, dest, obj))
        msg = bytes(f"{len(msg):<{self._header_size}}{self._password}", 'utf-8') + msg
        self._client.pending_acks[self._client.ack_stamp] = (time.monotonic(), msg)
        self.transport.write(msg)

    def send_ack(self, stamp):
        msg = self._client.serializer((stamp, 'ACK', None, None))
        msg = bytes(f"{len(msg):<{self._header_size}}{self._password}", 'utf-8') + msg
        self.transport.write(msg)

    def get_state(self):
        return self._state


class TLSClientFactory(ReconnectingClientFactory):
    protocol = ClientProtocol

    def __init__(self, client):
        self.password = client.password
        self._groups = client.groups
        self._header_size = client.header_size
        self._client = client
        self.maxDelay = self._client.recon_max_delay
        self.initialDelay = self._client.recon_initial_delay
        self.factor = self._client.recon_factor
        self.jitter = self._client.recon_jitter

    def startedConnecting(self, connector):
        logger.info('Started to connect.')

    def buildProtocol(self, addr):
        logger.info('Connected.')
        self.resetDelay()
        return ClientProtocol(client=self._client,
                              password=self.password,
                              groups=self._groups,
                              header_size=self._header_size)

    def clientConnectionLost(self, connector, reason):
        logger.info(f'Lost connection.  Reason: {reason.getErrorMessage()}')
        ReconnectingClientFactory.clientConnectionLost(self, connector, reason)

    def clientConnectionFailed(self, connector, reason):
        logger.info(f'Connection failed. Reason: {reason.getErrorMessage()}')
        ReconnectingClientFactory.clientConnectionFailed(self, connector, reason)


class Client:
    def __init__(self,
                 ip_server,
                 port_server,
                 password,
                 serializer,
                 deserializer,
                 header_size=10,
                 groups=None,
                 local_com_port=2097,
                 security="TLS",
                 keys_dir=None,
                 hostname='default',
                 recon_max_delay=60.0,
                 recon_initial_delay=10.0,
                 recon_factor=1.5,
                 recon_jitter=0.1):

        self.serializer = serializer
        self.deserializer = deserializer

        self.groups = groups
        self._ip_server = ip_server
        self._port_server = port_server
        self._local_com_port = local_com_port
        self.password = password
        self.header_size = header_size
        self._reactor = None
        self.to_server = None  # to communicate with the central relay
        self.endpoint = None  # to communicate with endpoint
        self.store = []
        self.ack_stamp = 0
        self.pending_acks = {}  # this contains copies of sent commands until corresponding ACKs are received
        self._security = security
        self._keys_dir = keys_dir
        self._hostname = hostname
        self.recon_max_delay = recon_max_delay
        self.recon_initial_delay = recon_initial_delay
        self.recon_factor = recon_factor
        self.recon_jitter = recon_jitter

    def run(self):
        """
        Main loop containing the reactor loop.

        This is run in its own process.
        """
        # from twisted.internet.interfaces import IReadDescriptor
        from twisted.internet import reactor

        # Initialize the local connection
        reactor.connectTCP(host='127.0.0.1', port=self._local_com_port, factory=LocalProtocolForClientFactory(self))

        # Initialize the Internet connection
        if self._security == "TCP":
            reactor.connectTCP(host=self._ip_server, port=self._port_server, factory=TLSClientFactory(client=self))
        elif self._security == "TLS":
            # Use default keys if none are provided
            self_signed = os.path.join(self._keys_dir, 'certificate.pem') if self._keys_dir is not None else os.path.join(get_default_keys_folder(), 'certificate.pem')
            # Authenticates the server to all potential clients for TLS communication
            try:
                certData = FilePath(self_signed).getContent()
            except OpenSSL.SSL.Error:
                raise AttributeError("The provided keys directory could not be found or does not contain the necessary keys. \
                    Make sure that you are providing a correct path, that your private key is named 'private.key' and that your public key is named 'selfsigned.crt'. \
                        You can use the script generate_certificates.py to generate the keys.")
            authority = ssl.Certificate.loadPEM(certData)            
            reactor.connectSSL(
                host=self._ip_server,
                port=self._port_server,
                factory=TLSClientFactory(client=self),
                contextFactory=ssl.optionsForClientTLS(hostname=self._hostname, trustRoot=authority)
            )
        else:
            logger.warning(f"Unsupported connection: {self._security}")
        
        # Start the reactor
        self._reactor = reactor
        reactor.run()

        # When done, deallocate reactor memory
        self._reactor = None

    def check_acks(self):
        """Returns true if we are not waiting for acknowledgements.

        Returns:
            bool: Whether the dictionary of pending acknowledgements is empty.
        """
        res = len(self.pending_acks.keys()) == 0 
        return res

    def close(self, counter):

        # Check if we are allowed to leave by looking at acknowledgements
        logger.debug(f"Attempting to terminate Endpoint for {counter}th time")

        if self.check_acks() or counter > 10:
            if self._reactor is not None:
                if self.to_server is not None:
                    self.to_server.transport.loseConnection()
                logger.info(f"Succesfully terminated endpoint connections")
                self._reactor.stop()
        else:
            from twisted.internet import reactor
            reactor.callLater(1, self.close, counter+1)


if __name__ == "__main__":
    cli = Client(ip_server="127.0.0.1", port_server=8123, password="pswd")
    cli.run()



# import signal
import queue


# try:
#     signal.signal(signal.SIGINT, signal.SIG_DFL)
# except Exception as e:
#     pass


def wait_event(event):
    """
    Workaround for an Event bug on Windows.

    See: https://bugs.python.org/issue35935
    """
    event.wait()


def get_from_queue(q, blocking=False):
    try:
        return [q.get(block=blocking)]
    except queue.Empty:
        return []



import unittest
import time

from utils import HelperTester


def same_lists_no_order(l1, l2):
    for elt in l1:
        if elt in l2:
            l2.remove(elt)
        else:
            return False
    return len(l2) == 0


class TestGroupsSync(unittest.TestCase):

    # Set up the server and all endpoints for all tests
    def setUp(self):
        self.ht = HelperTester(deserializer_mode="synchronous")

    def test_groups_accept_all(self):
        sr = self.ht.spawn_relay
        se = self.ht.spawn_endpoint
        relay = sr(accepted_groups=None)
        ep1 = se(groups='group1')
        ep2 = se(groups=('group1', 'group2'))
        ep3 = se(groups='group3')
        ep4 = se(groups='group5')
        ep5 = se(groups=('group6', 'group5', 'group1'))
        time.sleep(1.0)  # let everyone handshake the relay so that broadcasts don't get overwritten before that

        # test broadcasting

        ep5.send_object(obj='test1', destination='group1')
        r = ep1.pop(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test1', f"r:{r}")

        ep5.send_object(obj='test2', destination='group1')
        r = ep1.receive_all(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test2', f"r:{r}")

        ep2.send_object(obj='test3', destination='group1')
        r = ep1.receive_all(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test3', f"r:{r}")
        r = []
        while len(r) < 3:
            r += ep2.receive_all(blocking=True)
        self.assertEqual(len(r), 3, f"r:{r}")
        self.assertEqual(r[0], 'test1', f"r:{r}")
        self.assertEqual(r[1], 'test2', f"r:{r}")
        self.assertEqual(r[2], 'test3', f"r:{r}")

        ep1.send_object(obj='test4', destination=('group1', 'group5', 'group6'))
        r = ep1.pop(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test4', f"r:{r}")
        r = ep2.pop(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test4', f"r:{r}")
        r = ep4.pop(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test4', f"r:{r}")
        r = []
        while len(r) < 6:
            r += ep5.receive_all(blocking=True)
        self.assertEqual(len(r), 6, f"r:{r}")
        self.assertEqual(r[0], 'test1', f"r:{r}")
        self.assertEqual(r[1], 'test2', f"r:{r}")
        self.assertEqual(r[2], 'test3', f"r:{r}")
        self.assertEqual(r[3], 'test4', f"r:{r}")
        self.assertEqual(r[4], 'test4', f"r:{r}")
        self.assertEqual(r[5], 'test4', f"r:{r}")

        ep1.send_object(obj='test5', destination='group3')
        ep1.send_object(obj='test6', destination='group3')
        r = ep3.get_last(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        if r[0] == 'test5':
            r = ep3.get_last(blocking=True)
            self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test6', f"r:{r}")

        # test producing
        ep1.produce(obj='test7', group='group3')
        ep3.notify(groups='group3')
        r = ep3.pop(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test7', f"r:{r}")

        ep1.send_object(obj='test8', destination={'group1': 3})
        ep1.notify(groups='group1')
        ep2.notify(groups='group1')
        ep5.notify(groups='group1')
        r = ep1.pop(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test8', f"r:{r}")
        r = ep2.pop(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test8', f"r:{r}")
        r = ep5.pop(blocking=True)
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test8', f"r:{r}")

        # test mixed broadcasting / producing

        ep5.send_object(obj='test9', destination={'group1': 1, 'group2': 1})  # this produces to group 1 and 2
        ep5.send_object(obj='test10', destination={'group1': -1, 'group2': -1})  # this broadcasts to groups 1 and 2
        ep1.notify(groups=('group1', 'group2', 'group999'))  # note that ep1 is in group 1 only
        r = []
        while len(r) < 2:
            r += ep1.receive_all(blocking=True)
        self.assertEqual(len(r), 2, f"r:{r}")
        self.assertIn('test9', r, f"r:{r}")
        self.assertIn('test10', r, f"r:{r}")
        ep2.notify(groups=('group1', 'group2'))  # note that ep2 is in group 1 and group 2, but group 1 is empty
        r = []
        while len(r) < 3:
            r += ep2.receive_all(blocking=True)
        self.assertEqual(len(r), 3, f"r:{r}")
        self.assertTrue(same_lists_no_order(r, ['test9', 'test10', 'test10']), f"r:{r}")
        ep5.send_object(obj='test11', destination={'group1': 1})  # let us send one more consumable to group 1
        r = ep2.receive_all(blocking=True)  # now the notification sent by ep2 can be fulfilled
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test11', f"r:{r}")

        # test multiple producing / consuming

        ep5.send_object(obj='test12', destination={'group1': 10})  # this produces to group 1
        ep5.send_object(obj='test13', destination={'group2': 10})  # this produces to group 2
        ep2.notify(groups={'group1': 1})  # retrieve 1 elt in group 1
        r = []
        while len(r) < 11:  # one consumable in group1 and all consumables is group 2
            ep2.notify(groups={'group2': -1})  # ask for all elts in group 2
            r += ep2.receive_all(blocking=True)
        self.assertEqual(len(r), 11, f"r:{r}")
        self.assertTrue(same_lists_no_order(r, ['test12', ] + ['test13', ] * 10), f"r:{r}")
        r = []
        while len(r) < 9:  # all remaining consumables in group 1
            ep1.notify(groups={'group1': -1})  # ask for all elts in group 1
            r += ep1.receive_all(blocking=True)
        self.assertEqual(len(r), 9, f"r:{r}")
        self.assertTrue(same_lists_no_order(r, ['test12', ] * 9), f"r:{r}")

    def test_groups_accept_some(self):
        sr = self.ht.spawn_relay
        se = self.ht.spawn_endpoint
        accepted_groups = {
            'group1': {'max_count': 2, 'max_consumables': 2},
            'group2': {'max_count': 1, 'max_consumables': 1},
            'group3': {'max_count': 1, 'max_consumables': 0},
            'group4': {'max_count': None, 'max_consumables': None}
        }
        relay = sr(accepted_groups=accepted_groups)
        # nobody is in group 4, and groups 5 and 6 are not accepted
        ep1 = se(groups=('group6', 'group5', 'group1'))  # should not connect as group 5 and 6 are not allowed
        ep2 = se(groups='group1')  # should connect
        ep3 = se(groups=('group1', 'group2'))  # should connect
        ep4 = se(groups='group3')  # should connect
        time.sleep(1.0)  # let everyone connect so that old broadcasts are not lost for new clients
        ep5 = se(groups='group1')  # should not connect as group1 is full
        time.sleep(0.5)

        # test broadcasting

        ep5.send_object(obj='test1', destination='group1')  # should not send as ep5 is not connected
        # (the previous line should also output a warning)
        time.sleep(0.5)
        r = ep1.pop(blocking=False)  # not connected so should not receive
        self.assertEqual(len(r), 0, f"r:{r}")
        r = ep2.pop(blocking=False)  # should not receive since nothing should have been be sent
        self.assertEqual(len(r), 0, f"r:{r}")

        ep4.send_object(obj='test2', destination='group1')  # should send
        time.sleep(0.5)
        r = ep1.pop(blocking=False)  # not connected so should not receive
        self.assertEqual(len(r), 0, f"r:{r}")
        r = ep2.pop(blocking=True)  # should receive
        self.assertEqual(len(r), 1, f"r:{r}")
        self.assertEqual(r[0], 'test2')

        ep2.produce(obj='test3', group='group2')
        ep3.notify(groups='group2')
        r = []
        while len(r) < 2:
            r += ep3.receive_all(blocking=True)
        self.assertEqual(len(r), 2, f"r:{r}")
        self.assertTrue(same_lists_no_order(r, ['test2', 'test3']), f"r:{r}")

    def tearDown(self):
        self.ht.clear()


if __name__ == '__main__':
    unittest.main()



from setuptools import setup, find_packages
import sys


if sys.version_info < (3, 7):
    sys.exit('Sorry, Python < 3.7 is not supported, upgrade your python installation to use tlspyo.')


with open("README.md", "r") as fh:
    long_description = fh.read()


setup(name='tlspyo',
      packages=[package for package in find_packages()],
      version='0.3.0',
      download_url='https://github.com/MISTLab/tls-python-object/archive/refs/tags/v0.3.0.tar.gz',
      license='MIT',
      description='Secure transport of python objects using TLS encryption',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Yann Bouteiller, Milo Sobral',
      url='https://github.com/MISTLab/tls-python-object',
      keywords=['python', 'tls', 'ssl', 'pickle', 'transfer', 'object', 'transport', 'twisted'],
      install_requires=[
        'twisted',
        'pyOpenSSL>22.1.0',
        'service_identity',
        'platformdirs'
        ],
      extras_requires={
        "dev": ['pytest', 'pytest-timeout']
      },
      classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: Education',
          'Intended Audience :: Information Technology',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: MIT License',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Topic :: Utilities',
      ],
      )



from twisted.internet.protocol import Protocol, ClientFactory

from tlspyo.logs import logger


class LocalProtocolForServer(Protocol):

    def __init__(self, server):
        self._server = server
        self._state = "INIT"
        self._buffer = b""
        self._header_size = self._server.header_size
        self._identifier = None

    def connectionMade(self):
        assert self._state == "INIT", f"Bad state: {self._state}"
        self._state = "ALIVE"

    def connectionLost(self, reason):
        if self._server.has_client(self._identifier):
            self._server.delete_client(self._identifier)
        assert not self._server.has_client(self._identifier)
        self._identifier = None
        self._state = "DEAD"

    def dataReceived(self, data):
        try:
            self._buffer += data
            i = self._header_size
            if len(self._buffer) >= i:
                i, j = self.process_header()
                while len(self._buffer) >= j:
                    cmd, _ = self._server.deserializer(self._buffer[i:j])
                    if cmd == "STOP":
                        self.transport.loseConnection()
                        self._server.close(1)
                    elif cmd == 'TEST':
                        pass
                    else:
                        logger.warning(f"Local: Invalid command: {cmd}")
                        self._state = "CLOSED"
                        self.transport.abortConnection()
                    # truncate the processed part of the buffer:
                    self._buffer = self._buffer[j:]
                    i, j = self.process_header()
        except Exception as e:
            logger.warning(f"Local: Unhandled exception: {e}")
            self._state = "KILLED"
            self.transport.abortConnection()
            raise e

    def process_header(self):
        i = self._header_size
        if len(self._buffer) < i:
            return 0, len(self._buffer) + 1
        data_len = int(self._buffer[:self._header_size])
        j = i + data_len
        return i, j


class LocalProtocolForServerFactory(ClientFactory):
    protocol = LocalProtocolForServer

    def __init__(self, server):
        self.server = server

    def startedConnecting(self, connector):
        logger.info('Local for server: Started to connect.')

    def buildProtocol(self, addr):
        logger.info('Local for server: Connected.')
        return LocalProtocolForServer(self.server)

    def clientConnectionLost(self, connector, reason):
        logger.info(f'Local for server: lost connection.  Reason: {reason.getErrorMessage()}')

    def clientConnectionFailed(self, connector, reason):
        logger.info(f'Local for server: Client connection failed. Reason: {reason.getErrorMessage()}')
